// chunk.cpp
//
//
/*
    Copyright (C) 1999-2003 Jason Summers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    See the file readme.txt for more information.
*/


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <stdarg.h>
#include <assert.h>
#include "resource.h"
#include "tweakpng.h"
#ifdef TWPNG_HAVE_ZLIB
#include <zlib.h>
#endif

extern DWORD crc_table[256];
extern struct windowprefs_struct window_prefs;
extern struct globals_struct globals;

static char *known_text_keys[] = { "Title","Author","Description",
	"Copyright","Creation Time","Software","Disclaimer",
	"Warning","Source","Comment",NULL };


static BOOL CALLBACK DlgProcEdit_tEXt(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcEdit_PLTE(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcEditChunk(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcGetInt(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);


//#ifdef TWPNG_HAVE_ZLIB
//int uncompress_it(unsigned char **dataoutp, unsigned char *datain, int inlen);
//int compress_it(unsigned char **dataoutp, unsigned char *datain, int inlen);
//#endif

typedef struct {
	int value;   // initial value and final value
	char *label;
	char *title;  // window title
} get_int_struct_t;


typedef struct {
	unsigned char red;
	unsigned char green;
	unsigned char blue;
	unsigned char alpha;
} palent_t;

typedef struct {
	int numplte;
	int minplte;
	int maxplte;
	int numtrns;
	int mintrns;  // 1 if a trns chunk exists, otherwise 0
	int numbkgd;   // 1 if a bkgd chunk exists, otherwise 0
	palent_t  plte[256];
	unsigned char     bkgd;
	unsigned char     trnscolor; // only used for grayscale images

	HWND hwnd; // hack
	int bkgdbuttonstate;
	int alphabuttonstate;

	int caneditcolors; // can palette colors be edited? (not if grayscale)
}  palette_info_struct_t;


typedef struct {
	int id;
	char *name;
} chunk_id_struct_t;

static chunk_id_struct_t chunk_id_list[] = {
	{CHUNK_IHDR,"IHDR"},
	{CHUNK_IEND,"IEND"},
	{CHUNK_IHDR,"IHDR"},
	{CHUNK_IEND,"IEND"},
	{CHUNK_IDAT,"IDAT"},
	{CHUNK_PLTE,"PLTE"},
	{CHUNK_bKGD,"bKGD"},
	{CHUNK_cHRM,"cHRM"},
	{CHUNK_gAMA,"gAMA"},
	{CHUNK_hIST,"hIST"},
	{CHUNK_pHYs,"pHYs"},
	{CHUNK_sBIT,"sBIT"},
	{CHUNK_tEXt,"tEXt"},
	{CHUNK_tIME,"tIME"},
	{CHUNK_tRNS,"tRNS"},
	{CHUNK_zTXt,"zTXt"},
	{CHUNK_sRGB,"sRGB"},
	{CHUNK_iCCP,"iCCP"},
	{CHUNK_iTXt,"iTXt"},
	{CHUNK_sPLT,"sPLT"},
              
	{CHUNK_oFFs,"oFFs"},
	{CHUNK_pCAL,"pCAL"},
	{CHUNK_sCAL,"sCAL"},
	{CHUNK_gIFg,"gIFg"},
	{CHUNK_gIFx,"gIFx"},
	{CHUNK_gIFt,"gIFt"},
	{CHUNK_fRAc,"fRAc"},
              
	{CHUNK_MHDR,"MHDR"},
	{CHUNK_MEND,"MEND"},
          
	{CHUNK_LOOP,"LOOP"},
	{CHUNK_ENDL,"ENDL"},
	{CHUNK_DEFI,"DEFI"},
	{CHUNK_JHDR,"JHDR"},
	{CHUNK_BASI,"BASI"},
	{CHUNK_CLON,"CLON"},
	{CHUNK_DHDR,"DHDR"},
	{CHUNK_PAST,"PAST"},
	{CHUNK_DISC,"DISC"},
	{CHUNK_BACK,"BACK"},
	{CHUNK_FRAM,"FRAM"},
	{CHUNK_MOVE,"MOVE"},
	{CHUNK_CLIP,"CLIP"},
	{CHUNK_SHOW,"SHOW"},
	{CHUNK_TERM,"TERM"},
	{CHUNK_SAVE,"SAVE"},
	{CHUNK_SEEK,"SEEK"},
	{CHUNK_eXPI,"eXPI"},
	{CHUNK_fPRI,"fPRI"},
	{CHUNK_nEED,"nEED"},
	{CHUNK_pHYg,"pHYg"},
	{CHUNK_JDAT,"JDAT"},
	{CHUNK_JSEP,"JSEP"},
	{CHUNK_PROM,"PROM"},
	{CHUNK_IPNG,"IPNG"},
	{CHUNK_PPLT,"PPLT"},
	{CHUNK_IJNG,"IJNG"},
	{CHUNK_DROP,"DROP"},
	{CHUNK_DBYK,"DBYK"},
	{CHUNK_ORDR,"ORDR"},
	{0,NULL}};


int Chunk::is_critical(void)      { return (chunktype[0]&0x20)?0:1; }
int Chunk::is_public(void)        { return (chunktype[1]&0x20)?0:1; }
int Chunk::is_safe_to_copy(void)  { return (chunktype[3]&0x20)?1:0; }


#ifdef TWPNG_HAVE_ZLIB

static DWORD get_uncompressed_size(unsigned char *datain, int inlen)
{
	static int blksize = 16384;
	unsigned char dbuf[16384];
	int data_read=0;
	int err;

	z_stream z;

	ZeroMemory((void*)&z,sizeof(z_stream));

	z.opaque=0;
	z.next_in = datain;
	z.avail_in = inlen;

	inflateInit(&z);

	while(1) {
		z.next_out = dbuf;
		z.avail_out = blksize;
		err = inflate(&z, Z_NO_FLUSH);
		if(err == Z_STREAM_END) break;
		if(err != Z_OK) {
			inflateEnd(&z);
			return 0;
		}
	}
	inflateEnd(&z);
	return z.total_out;
}


// returns length of compressed data
// allocs a new buffer for the data
static int compress_it(unsigned char **dataoutp, unsigned char*datain, int inlen)
{
	z_stream z;
	unsigned char *dataout;
	int alloced;
	int err;

	alloced=  (int)((double)inlen*1.02) + 50;
	dataout= (unsigned char*)malloc(alloced);
	if(!dataout) {
		mesg(MSG_E,"can't alloc memory for compress");
		return 0;
	}

	ZeroMemory((void*)&z,sizeof(z_stream));

	z.opaque=0;
	z.next_in = datain;
	z.avail_in = inlen;
	z.next_out =  dataout;
	z.avail_out = alloced;

	err = deflateInit(&z, globals.compression_level);   //Z_DEFAULT_COMPRESSION

	err = deflate(&z, Z_FINISH);
	if(err != Z_STREAM_END) {
		mesg(MSG_E,"error compressing data");
		err = deflateEnd(&z);
		(*dataoutp)=NULL;
		return 0;
	}

	err = deflateEnd(&z);
	(*dataoutp)=dataout;
	return z.total_out;
}

static int uncompress_it(unsigned char **dataoutp, unsigned char*datain, int inlen)
{
	int alloced=0;
	unsigned char *dataout;
	z_stream z;
	int data_written;
	int ret;
	DWORD unc_size;

	// We uncompress it once just to figure out how big a buffer we
	// need to allocate. It's wasteful, but no big deal
	// in this program, I think.
	unc_size=get_uncompressed_size(datain,inlen);
	if(!unc_size) {
		mesg(MSG_E,"inflate error");
		return 0;
	}
	assert((*dataoutp)==NULL);

	alloced=unc_size+50;       // just unc_size should work

	dataout= (unsigned char*)malloc(alloced);
	if(!dataout) {
		mesg(MSG_E,"can't alloc memory for uncompress");
		return 0;
	}

	data_written=0;

	ZeroMemory((void*)&z,sizeof(z_stream));

	z.opaque=0;
	z.next_in = datain;
	z.avail_in = inlen;
	z.next_out =  dataout;         // &dataout[data_written];
	z.avail_out = alloced;

	inflateInit(&z);

	ret = inflate(&z,Z_FINISH);
	if(ret != Z_STREAM_END) {
		mesg(MSG_E,"decompression error");
	}

	inflateEnd(&z);

	(*dataoutp)=dataout;
	return z.total_out;
}
#endif

// returns 0 if canceled
int choose_color_dialog(HWND hwnd, unsigned char *redp,
						unsigned char *greenp, unsigned char *bluep)
{
	//static int hack=0;
	//static COLORREF custcolors[16];
	CHOOSECOLOR cc;

	//if(!hack) {
	//	ZeroMemory((void*)custcolors,16*sizeof(COLORREF));
	//	hack=1;
	//}

	cc.lStructSize=sizeof(CHOOSECOLOR);
	cc.hwndOwner=hwnd;
	cc.hInstance=NULL;
	cc.rgbResult= RGB(*redp,*greenp,*bluep);  // set initial value
	cc.lpCustColors=globals.custcolors;
	cc.Flags=CC_RGBINIT|CC_FULLOPEN;
	cc.lCustData=0;
	cc.lpfnHook=NULL;
	cc.lpTemplateName=NULL;

	if(ChooseColor(&cc)) {
		(*redp)   = GetRValue(cc.rgbResult);
		(*greenp) = GetGValue(cc.rgbResult);
		(*bluep)  = GetBValue(cc.rgbResult);
		return 1;
	}
	return 0;
}


// indata is not null terminated,
// outdata is null terminated, and must be freed by caller
static void lf2crlf(unsigned char **outdatap, unsigned char *indata, int len)
{
	int lfcount=0;
	unsigned char *outdata;
	int i,p;

	(*outdatap)=NULL;

	for(i=0;i<len;i++) {
		if(indata[i]=='\n') lfcount++;
	}

	outdata=(unsigned char*)malloc(len+lfcount+1);
	if(!outdata) return;

	p=0;
	for(i=0;i<len;i++) {
		if(indata[i]=='\n') {
			outdata[p++]='\r';
			outdata[p++]='\n';
		}
		else {
			outdata[p++]=indata[i];
		}
	}
	assert(p==len+lfcount);
	outdata[p]='\0';

	(*outdatap)=outdata;
}

// input must be null-terminated
// output will be null-terminated
// this conversion is in-place
static void crlf2lf(unsigned char *outdata)
{
	int d,i,len;

	len=strlen((char*)outdata);

	d=0;
	for(i=0;i<len;i++) {
		if(outdata[i]=='\r' && outdata[i+1]=='\n') {
			outdata[i-d]='\n';
			d++;
		}
		else {
			outdata[i-d]=outdata[i];
		}
	}
	outdata[len-d]='\0';
}

int Chunk::copy_to_memory(unsigned char *m)
{
	write_int32(&m[0],length);
	memcpy(&m[4],chunktype,4);
	memcpy(&m[8],data,length);
	write_int32(&m[8+length],crc);
	return 1;
}

// returns bytes consumed; 0 if error
int Chunk::init_from_memory(unsigned char *m, int msize)
{
	if(msize<12) return 0;

	length= read_int32(&m[0]);
	if((int)length+12>msize) return 0;

	memcpy(chunktype,&m[4],4);
	chunktype[4]='\0';

	data=(unsigned char*)malloc(length);
	if(!data) {
		mesg(MSG_S,"can't alloc memory for new chunk");
		return 0;
	}
	memcpy(data,&m[8],length);
	// crc is next, but we'll ignore it

	after_init();
	chunkmodified();

	return length+12;
}

int Chunk::edit_plte_info(void)
{
	Chunk *ch_plte;
	Chunk *ch_trns;
	Chunk *ch_bkgd;
	palette_info_struct_t pal_info;
	int maxpal;
	int i;
	int changed;
	int grayscale;

	// set defaults

	ch_plte=ch_trns=ch_bkgd=NULL;
	grayscale= (parentpng->colortype==0 || parentpng->colortype==4);

	for(i=0;i<256;i++) {
		pal_info.plte[i].red   = 0;
		pal_info.plte[i].green = 0;
		pal_info.plte[i].blue  = 0;
		pal_info.plte[i].alpha = 255;
	}
	pal_info.numbkgd=0;
	pal_info.bkgd=0;
	pal_info.numtrns=0;
	pal_info.mintrns=0;
	pal_info.minplte=1;
	pal_info.bkgdbuttonstate=0;
	pal_info.alphabuttonstate=0;
	pal_info.caneditcolors=1;
	pal_info.trnscolor=0;

	maxpal= 1<<parentpng->bitdepth;

	switch(chunktype_id) {
	case CHUNK_PLTE:
		ch_plte=this;
		break;
	case CHUNK_tRNS:
		ch_trns=this;
		pal_info.alphabuttonstate=1;
		break;
	case CHUNK_bKGD:
		ch_bkgd=this;
		pal_info.bkgdbuttonstate=1;
		break;
	}

	if(!ch_plte && !grayscale) ch_plte= parentpng->find_first_chunk(CHUNK_PLTE,NULL);
	if(!ch_trns) ch_trns= parentpng->find_first_chunk(CHUNK_tRNS,NULL);
	if(!ch_bkgd) ch_bkgd= parentpng->find_first_chunk(CHUNK_bKGD,NULL);

	if(!ch_plte && !grayscale) {
		mesg(MSG_E,"Need a PLTE chunk to edit this");
		return 0;
	}

	if(ch_trns) {
		pal_info.mintrns= 1;
		if(grayscale) {
			pal_info.numtrns=1;
			if(ch_trns->length==2) {
				pal_info.trnscolor = read_int16(&ch_trns->data[0]);
			}
		}
		else {
			pal_info.numtrns= ch_trns->length;
			if(pal_info.numtrns > maxpal) pal_info.numtrns= maxpal;

			for(i=0;i<pal_info.numtrns;i++) {
				pal_info.plte[i].alpha= ch_trns->data[i];
			}
		}
	}

	if(ch_bkgd) {
		if(grayscale) {
			if(ch_bkgd->length == 2) {
				pal_info.numbkgd= 1;
				pal_info.bkgd = read_int16(&ch_bkgd->data[0]);
			}
		}
		else {
			if(ch_bkgd->length == 1) {
				pal_info.numbkgd= 1;
				pal_info.bkgd= ch_bkgd->data[0];
			}
		}
	}

	if(grayscale) {
		pal_info.maxplte= maxpal;
		pal_info.numplte= maxpal;
		for(i=0;i<maxpal;i++) {
			pal_info.plte[i].red = i*(255/(maxpal-1));
			pal_info.plte[i].green = pal_info.plte[i].red;
			pal_info.plte[i].blue  = pal_info.plte[i].red;
		}
		pal_info.caneditcolors=0;
	}
	else {
		pal_info.maxplte= maxpal;
		pal_info.numplte= (ch_plte->length)/3;
		if(pal_info.numplte > maxpal) pal_info.numplte= maxpal;
		for(i=0;i<pal_info.numplte;i++) {
			pal_info.plte[i].red   = ch_plte->data[3*i];
			pal_info.plte[i].green = ch_plte->data[3*i+1];
			pal_info.plte[i].blue  = ch_plte->data[3*i+2];
		}
	}
	changed=DialogBoxParam(globals.hInst,"DLG_PLTE",globals.hwndMain,
		(DLGPROC)DlgProcEdit_PLTE, (LPARAM)(&pal_info));
	if(changed>0) {
		if(ch_bkgd) {
			if(grayscale) {
				write_int16(&ch_bkgd->data[0],pal_info.bkgd);
			}
			else {
				ch_bkgd->data[0]= pal_info.bkgd;
			}
			ch_bkgd->chunkmodified();
		}
		if(ch_trns) {
			if(grayscale) {
				if(ch_trns->length==2) {
					write_int16(&ch_trns->data[0],pal_info.trnscolor);
				}
			}
			else {
				if(ch_trns->length != (unsigned int)pal_info.numtrns) {
					// size of alpha palette changed, need to reallocate
					free(ch_trns->data);
					ch_trns->data=(unsigned char*)malloc(pal_info.numtrns);
					ch_trns->length=(unsigned int)pal_info.numtrns;
				}
				for(i=0;i<pal_info.numtrns;i++) {
					ch_trns->data[i]=pal_info.plte[i].alpha;
				}
			}
			ch_trns->chunkmodified();
		}
		if(ch_plte) {
			if(ch_plte->length != (unsigned int)(3*pal_info.numplte) ) {
				// size of palette changed, need to reallocate
				free(ch_plte->data);
				ch_plte->data=(unsigned char*)malloc(3*pal_info.numplte);
				ch_plte->length=(unsigned int)(3*pal_info.numplte);
			}
			for(i=0;i<pal_info.numplte;i++) {
				ch_plte->data[3*i  ]=pal_info.plte[i].red;
				ch_plte->data[3*i+1]=pal_info.plte[i].green;
				ch_plte->data[3*i+2]=pal_info.plte[i].blue;
			}
			ch_plte->chunkmodified();
		}
	}
	if(changed==1) changed=2;
	return changed;
}


// return 1 if changed, 2 if changed multiple chunks
int Chunk::edit(void)
{
	int changed=0;

	switch(chunktype_id) {
	case CHUNK_tEXt:
#ifdef TWPNG_HAVE_ZLIB
	case CHUNK_zTXt:
#endif
		changed=  DialogBoxParam(globals.hInst,"DLG_TEXT",globals.hwndMain,
		    (DLGPROC)DlgProcEdit_tEXt, (LPARAM)this);
		break;

	case CHUNK_IHDR:
		changed=  DialogBoxParam(globals.hInst,"DLG_IHDR",globals.hwndMain,
		    (DLGPROC)DlgProcEditChunk, (LPARAM)this);
		if(changed>0) { after_init(); changed=2; }
		break;
	case CHUNK_gAMA:
		changed=  DialogBoxParam(globals.hInst,"DLG_GAMA",globals.hwndMain,
		    (DLGPROC)DlgProcEditChunk, (LPARAM)this);
		break;
	case CHUNK_sRGB:
		changed=  DialogBoxParam(globals.hInst,"DLG_SRGB",globals.hwndMain,
		    (DLGPROC)DlgProcEditChunk, (LPARAM)this);
		break;
	case CHUNK_pHYs:
		changed=  DialogBoxParam(globals.hInst,"DLG_PHYS",globals.hwndMain,
		    (DLGPROC)DlgProcEditChunk, (LPARAM)this);
		break;
	case CHUNK_cHRM:
		changed=  DialogBoxParam(globals.hInst,"DLG_CHRM",globals.hwndMain,
		    (DLGPROC)DlgProcEditChunk, (LPARAM)this);
		break;
	case CHUNK_bKGD:
		switch(parentpng->colortype) {
		case 3:         // color palette
			goto plte_start;

		case 0: case 4:  // grayscale
			if(parentpng->bitdepth<=8) goto plte_start;
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR1",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		case 2: case 6:  // rgb
			if(parentpng->bitdepth==8) {
				unsigned char cr,cg,cb;
				cr=data[1];
				cg=data[3];
				cb=data[5];

				if(choose_color_dialog(globals.hwndMain, &cr,&cg,&cb) ) {
					data[0]=0; data[1]= cr;
					data[2]=0; data[3]= cg;
					data[4]=0; data[5]= cb;
					changed=1;
				}
			}
			else {
				changed=  DialogBoxParam(globals.hInst,"DLG_COLOR3",globals.hwndMain,
					(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			}
			break;
		}
		break;

	case CHUNK_sBIT:
		switch(parentpng->colortype) {
		case 0:   // grayscale
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR1",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		case 2: case 3:  // rbg or paletted
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR3",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		case 4:  // grayscale+alpha
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR2",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		case 6:  // rgba
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR4",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		}
		break;

	case CHUNK_tRNS:
		switch(parentpng->colortype) {
		case 3:
			goto plte_start;
		case 0: case 4:  // grayscale
			if(parentpng->colortype==0 && parentpng->bitdepth<=8) goto plte_start;
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR1",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		case 2: case 6:  // rgb
			changed=  DialogBoxParam(globals.hInst,"DLG_COLOR3",globals.hwndMain,
				(DLGPROC)DlgProcEditChunk, (LPARAM)this);
			break;
		}
		break;

	case CHUNK_PLTE:
plte_start:
		changed = edit_plte_info();
		break;

	case CHUNK_tIME:
		changed=  DialogBoxParam(globals.hInst,"DLG_TIME",globals.hwndMain,
		    (DLGPROC)DlgProcEditChunk, (LPARAM)this);
		break;

	}

	if(changed>0) chunkmodified();    // update crc
	//return (changed>0);
	return changed;
}

// call after you've changed a chunk
void Chunk::chunkmodified(void)
{
	crc=calc_crc();
}

DWORD Chunk::calc_crc(void)
{
	DWORD ccrc;  // calculated crc

	ccrc=update_crc(CRCINIT,(unsigned char*)chunktype,4);
	ccrc=update_crc(ccrc,data,length);
	ccrc=CRCCOMPL(ccrc);
	return ccrc;
}

int Chunk::get_chunk_type_id(void)
{
	int i;
	for(i=0; chunk_id_list[i].id; i++) {
		if(!strcmp(chunktype,chunk_id_list[i].name))
			return chunk_id_list[i].id;
	}

	return CHUNK_UNKNOWN;
}

// returns 1 if found, 0 if not found; caller must provide name[5]
int get_name_from_id(char *name, int x)
{
	int i;
	for(i=0; chunk_id_list[i].id; i++) {
		if(chunk_id_list[i].id==x) {
			strcpy(name,chunk_id_list[i].name);
			return 1;
		}
	}
	strcpy(name,"????");
	return 0;
}

void Chunk::describe_IHDR(char *buf)
{
	char buf2[500];
	if(length!=13) {
		strcpy(buf,"invalid PNG image header");
		return;
	}
	sprintf(buf,"PNG image header: %ux%u",read_int32(&data[0]),read_int32(&data[4]));
	sprintf(buf2,", %u bit%s/%s",(DWORD)(data[8]),(data[8]==1)?"":"s",
		(data[9]==3)?"pixel":"sample");
	strcat(buf,buf2);

	switch(data[9]) {
	case 0: sprintf(buf2,", grayscale"); break;
	case 2: sprintf(buf2,", truecolor"); break;
	case 3: sprintf(buf2,", paletted"); break;
	case 4: sprintf(buf2,", grayscale+alpha"); break;
	case 6: sprintf(buf2,", truecolor+alpha"); break;
	default: sprintf(buf2,", illegal color type (%u)",(DWORD)(data[9]));
	}
	strcat(buf,buf2);

	switch(data[12]) {
	case 0: sprintf(buf2,", noninterlaced"); break;
	case 1: sprintf(buf2,", interlaced"); break;
	default:
		sprintf(buf2,", unrecognized interlace method (%u)",(DWORD)(data[12]));
	}
	strcat(buf,buf2);
}


void Chunk::describe_MHDR(char *buf)
{
	if(length!=28) {
		sprintf(buf,"invalid MNG file header");
		return;
	}
	sprintf(buf,"MNG file header: %ux%u",
	(DWORD)read_int32(&data[0]),
	(DWORD)read_int32(&data[4]) );

}


void Chunk::describe_IEND(char *buf)
{
//	if(parentpng->imgtype==IMG_PNG)
//		strcpy(buf,"end-of-file marker");
//	else
		strcpy(buf,"end-of-image marker");
}

void Chunk::describe_bKGD(char *buf)
{
	if(parentpng->imgtype != IMG_PNG) {
		sprintf(buf,"background color");
		return;
	}

	if(parentpng->colortype==3) {   // paletted
		if(length==1) {
			sprintf(buf,"background color = palette entry %u",(DWORD)data[0]);
			return;
		}
	}
	else if(parentpng->colortype==0 || parentpng->colortype==4) { // grayscale
		if(length==2) {
			sprintf(buf,"background color = (%u)",
				(DWORD)read_int16(&data[0]));
			return;
		}
	}
	else if(parentpng->colortype==2 || parentpng->colortype==6) {  // truecolor
		if(length==6) {
			sprintf(buf,"background color = (%u,%u,%u)",
				(DWORD)read_int16(&data[0]),
				(DWORD)read_int16(&data[2]),
				(DWORD)read_int16(&data[4]));
			return;
		}
	}

	sprintf(buf,"invalid background color chunk");

}

void Chunk::describe_hIST(char *buf)
{
	if(length<2 || length>512 || (length%2)) {
		sprintf(buf,"invalid histogram chunk");
		return;
	}
	sprintf(buf,"histogram for palette colors, %u entr%s",length/2,(length==2)?"y":"ies");
}

void Chunk::describe_PLTE(char *buf)
{
	char buf2[80];

	switch(parentpng->colortype) {
	case 2: case 6:
		sprintf(buf,"palette (suggested)"); break;
	case 0: case 4:
		sprintf(buf,"illegal palette chunk"); break;
	default:	
		sprintf(buf,"palette");
	}

	if(length<3 || length > 768 || (length%3 !=0)) {
		strcat(buf,", incorrect palette size");
		return;
	}
	sprintf(buf2,", %u entr%s",(DWORD)(length/3),(length==3)?"y":"ies");
	strcat(buf,buf2);
}

void Chunk::describe_tRNS(char *buf)
{
	switch(parentpng->colortype) {
	case 3:
		if(length<=256) {
			sprintf(buf,"alpha values for palette colors, %u entr%s",length,(length==1)?"y":"ies");
			return;
		}
	case 0:
		if(length==2) {
			sprintf(buf,"transparent color = (%u)",(DWORD)read_int16(&data[0]));
			return;
		}
		break;
	case 2:
		if(length==6) {
			sprintf(buf,"transparent color = (%u,%u,%u)",
				(DWORD)read_int16(&data[0]),
				(DWORD)read_int16(&data[2]),
				(DWORD)read_int16(&data[4]));
			return;
		}
		break;
	default:
		sprintf(buf,"illegal transparency chunk, alpha channel is present");
		return;
	}
	sprintf(buf,"incorrect chunk length");
}

void Chunk::describe_sRGB(char *buf)
{
	char s[40];

	if(length!=1) {
		sprintf(buf,"incorrect chunk length");
		return;
	}
	switch(data[0]) {
	case 0: strcpy(s,"Perceptual"); break;
	case 1: strcpy(s,"Relative colorimetric"); break;
	case 2: strcpy(s,"Saturation"); break;
	case 3: strcpy(s,"Absolute colorimetric"); break;
	default: sprintf(s,"[unknown: %d]",(int)data[0]);
	}
	sprintf(buf,"sRGB color space, rendering intent: %s",s);
}

void Chunk::describe_gAMA(char *buf)
{
	DWORD g;
	double g2;

	if(length!=4) {
		sprintf(buf,"incorrect chunk length");
		return;
	}

	g=read_int32(&data[0]);
	g2=((double)(g))/100000.0;
	sprintf(buf,"file gamma = %.5f",g2);

}

void Chunk::describe_sBIT(char *buf)
{
	switch(parentpng->colortype) {
	case 0:
		if(length==1) {
			sprintf(buf,"significant bits/sample: %u",(DWORD)data[0]);
			return;
		}
		break;
	case 2: case 3:
		if(length==3) {
			sprintf(buf,"significant bits/sample: R:%u,G:%u,B:%u",
				(DWORD)data[0],(DWORD)data[1],(DWORD)data[2]);
			return;
		}
		break;
	case 4:
		if(length==2) {
			sprintf(buf,"significant bits/sample: %u,A:%u",
				(DWORD)data[0],(DWORD)data[1]);
			return;
		}
		break;
	case 6:
		if(length==4) {
			sprintf(buf,"significant bits/sample: R:%u,G:%u,B:%u,A:%u",
				(DWORD)data[0],(DWORD)data[1],(DWORD)data[2],(DWORD)data[3]);
			return;
		}
	}

	sprintf(buf,"incorrect chunk length");
}

void Chunk::describe_cHRM(char *buf)
{
	if(length!=32) {
		sprintf(buf,"incorrect chunk length");
		return;
	}
	sprintf(buf,"chromaticities: WP(%.5f,%.5f),R(%.5f,%.5f),G(%.5f,%.5f),B(%.5f,%.5f)",
		((double)read_int32(&data[0]))/100000.0,
		((double)read_int32(&data[4]))/100000.0,
		((double)read_int32(&data[8]))/100000.0,
		((double)read_int32(&data[12]))/100000.0,
		((double)read_int32(&data[16]))/100000.0,
		((double)read_int32(&data[20]))/100000.0,
		((double)read_int32(&data[24]))/100000.0,
		((double)read_int32(&data[28]))/100000.0);
}

void Chunk::describe_pHYs(char *buf)
{
	char s[80];
	int x,y;

	if(length!=9) {
		sprintf(buf,"incorrect chunk length");
		return;
	}

	x=read_int32(&data[0]);
	y=read_int32(&data[4]);

	sprintf(buf,"pixel size = %ux%u pixels",x,y);
	switch(data[8]) {
	case 0: strcat(buf," (per unspecified unit)"); break;
	case 1:
		sprintf(s," per meter (%.1fx%.1f dpi)",
			(double)x * 0.0254,
			(double)y * 0.0254);
		strcat(buf,s);
		break;
	default: strcat(buf," per (unrecognized unit)");
	}
}

// keyword and indata must be null-terminated
int Chunk::set_text_info(int newchunktype_id,unsigned char *keyword,unsigned char *indata)
{
	unsigned char *u;
	int kwlen,len;

#ifndef TWPNG_HAVE_ZLIB
	if(newchunktype_id==CHUNK_zTXt) {
		mesg(MSG_S,"internal: can't modify zTXt chunk, not zlib support");
		return 0;
	}
#endif

	if(uncompressed_data) {
		free((void*)uncompressed_data);
	}

	uncompressed_data_size=strlen((char*)indata);
	uncompressed_data=(unsigned char*)malloc(uncompressed_data_size);
	if(!uncompressed_data) return 0;

	// note uncompressed_data is not null-terminated
	memcpy((void*)uncompressed_data,(void*)indata,uncompressed_data_size);

	kwlen=strlen((char*)keyword);

	if(data) { free(data); data=NULL; } // lose the old data

	if(newchunktype_id==CHUNK_zTXt) {
#ifdef TWPNG_HAVE_ZLIB
		len=compress_it(&u,uncompressed_data,uncompressed_data_size);
		if(len<1 || !u) { return 0; }
		length=kwlen+2+len;
		data=(unsigned char*)malloc(length);
		if(!data) return 0;
		memcpy((void*)data,(void*)keyword,kwlen);
		data[kwlen]='\0';    // null separator
		data[kwlen+1]=0;   // compression type
		memcpy((void*)&data[kwlen+2],(void*)u,len);  // the compressed text
		strcpy(chunktype,"zTXt");
		chunktype_id=CHUNK_zTXt;
		free(u);
#endif
	}
	else {    // uncompressed text chunk
		length=kwlen+1+uncompressed_data_size;
		data=(unsigned char*)malloc(length);
		if(!data) return 0;
		memcpy((void*)data,(void*)keyword,kwlen);
		data[kwlen]='\0';    // null separator
		memcpy((void*)&data[kwlen+1],(void*)uncompressed_data,uncompressed_data_size);
		strcpy(chunktype,"tEXt");
		chunktype_id=CHUNK_tEXt;
	}
	return 1;
}

// returns data in a tEXt or zTXt chunk
// the pointers returned should NOT be freed by the caller
// the text will be in uncompressed_data. it is not NULL-terminated
int Chunk::get_text_info(unsigned char **keywordp)
{
	static unsigned char keyword[80];
	int p,i;
	unsigned char method;

	// find null character that marks the end of the keyword
	p= -1;
	for(i=0;i<(int)length;i++) {
		if(data[i]==0) { p=i; break;}
	}

	// p is now the length of the keyword
	if(p<1 || p>79) {
		return 0;
	}
	strncpy((char*)keyword,(char*)(&data[0]),p+1);
	(*keywordp)=keyword;

	if(chunktype_id==CHUNK_zTXt) {
		method= data[p+1];
		if(method!=0) {    // unknown compression method: force failure
			uncompressed_data_tried=1;
			uncompressed_data=NULL;
			return 0;
		}

#ifdef TWPNG_HAVE_ZLIB

		if(!uncompressed_data_tried) {   // if we haven't tried to uncompress this yet...
			uncompressed_data_tried=1;

			uncompressed_data_size=uncompress_it(&uncompressed_data, &data[p+2], length-p-2);
		}
		if(uncompressed_data_size<1) {    // failed
			if(uncompressed_data) free(uncompressed_data);
			uncompressed_data=NULL;
			return 0;
		}

#else
		// no zlib support, can only find keyword
		uncompressed_data=NULL;
		uncompressed_data_size=0;
		return 1;
#endif	
	}
	else {    // uncompressed text
		if(!uncompressed_data_tried) {
			uncompressed_data_tried=1;
			uncompressed_data= (unsigned char*)malloc(length-(p+1));
			if(!uncompressed_data) {
				mesg(MSG_S,"can't alloc memory");
				return 0;
			}
			uncompressed_data_size=length-(p+1);
			memcpy(uncompressed_data,&data[p+1],uncompressed_data_size);
		}
	}

	return 1;
}

// how much text will we display in the listbox
#define MAX_TEXT_DISPLAY 150

void Chunk::describe_text(char *buf, int ct)
{
	unsigned char *keyword;
	unsigned char text[MAX_TEXT_DISPLAY+10];
	unsigned int i;
	char *stext,*ctext;
	int flag;

	if(!get_text_info(&keyword)) {
		sprintf(buf,"can't read text chunk");
		return;
	}

	ctext= (ct==CHUNK_zTXt)?"compressed ":"";

	if(!uncompressed_data) {       // failed
		if(ct==CHUNK_zTXt) {
#ifdef TWPNG_HAVE_ZLIB
			sprintf((char*)text," --decompression failed!--");
#else
			sprintf((char*)text," --decompression not supported--");
#endif
		}
		else {         // uncompressed text chunk
			sprintf((char*)text," --can't read text--");
		}
	}
	else {
		text[0]='[';
		if(uncompressed_data_size<=MAX_TEXT_DISPLAY) {
			memcpy(&text[1],uncompressed_data,uncompressed_data_size);
			text[uncompressed_data_size+1]=']';
			text[uncompressed_data_size+2]='\0';
		}
		else {
			memcpy(&text[1],uncompressed_data,MAX_TEXT_DISPLAY);
			text[MAX_TEXT_DISPLAY+1]='.';
			text[MAX_TEXT_DISPLAY+2]='.';
			text[MAX_TEXT_DISPLAY+3]='.';
			text[MAX_TEXT_DISPLAY+4]='\0';
		}
	}

	// clobber bad characters
	for(i=0;keyword[i];i++) {
		if(keyword[i]<32 || (keyword[i]>126 && keyword[i]<161))
			keyword[i]='|';
	}
	for(i=0;text[i];i++) {
		if(text[i]<32 || (text[i]>126 && text[i]<161))
			text[i]='|';
	}

	flag=0;
	for(i=0; known_text_keys[i]; i++) {
		if(!strcmp((char*)keyword,known_text_keys[i])) flag=1;
	}
	stext= flag?"standard ":"nonstandard ";
	sprintf(buf,"%s%stext: [%s]=%s",stext,ctext,keyword,text);

}

static char *monthname[] = { "INVALID-MONTH","Jan","Feb","Mar","Apr","May",
		"Jun","Jul","Aug","Sep","Oct","Nov","Dec" };

void Chunk::describe_tIME(char *buf)
{

	int year, month, day, hour, minute, second;
	if(length!=7) {
		sprintf(buf,"invalid last modification time");
		return;
	}

	year=   read_int16(&data[0]);
	month=  (int)data[2];
	day=    (int)data[3];
	hour=   (int)data[4];
	minute= (int)data[5];
	second= (int)data[6];

	if(month<1 || month>12) month=0;

	sprintf(buf,"time of last modification = %d %s %d, %02d:%02d:%02d UTC",
		day,monthname[month],year,hour,minute,second);
}

void Chunk::get_text_descr(char *buf)
{
	strcpy(buf,"");

	if(!is_public()) {
		strcpy(buf,"private chunk");
		return;
	}

	switch(chunktype_id) {
	case CHUNK_IHDR: describe_IHDR(buf); break;
	case CHUNK_IEND: describe_IEND(buf); break;
	case CHUNK_bKGD: describe_bKGD(buf); break;
	case CHUNK_PLTE: describe_PLTE(buf); break;
	case CHUNK_tRNS: describe_tRNS(buf); break;
	case CHUNK_sBIT: describe_sBIT(buf); break;
	case CHUNK_gAMA: describe_gAMA(buf); break;
	case CHUNK_cHRM: describe_cHRM(buf); break;
	case CHUNK_pHYs: describe_pHYs(buf); break;
	case CHUNK_tIME: describe_tIME(buf); break;
	case CHUNK_hIST: describe_hIST(buf); break;
	case CHUNK_tEXt: describe_text(buf,CHUNK_tEXt); break;
	case CHUNK_zTXt: describe_text(buf,CHUNK_zTXt); break;

	case CHUNK_IDAT: strcpy(buf,"PNG image data"); break;

	case CHUNK_sRGB: describe_sRGB(buf); break;
	case CHUNK_iCCP: describe_keyword_chunk(buf,CHUNK_iCCP); break; 
	case CHUNK_iTXt: describe_keyword_chunk(buf,CHUNK_iTXt); break;
	case CHUNK_sPLT: describe_keyword_chunk(buf,CHUNK_sPLT); break;

	case CHUNK_oFFs: strcpy(buf,"image offset"); break;
	case CHUNK_pCAL: strcpy(buf,"calibration of pixel values"); break;
	case CHUNK_sCAL: strcpy(buf,"physical scale of image subject"); break;
	case CHUNK_gIFg: strcpy(buf,"GIF graphic control extension"); break;
	case CHUNK_gIFx: strcpy(buf,"GIF application extension"); break;
	case CHUNK_gIFt: strcpy(buf,"GIF plain text extension [DEPRECATED]"); break;
	case CHUNK_fRAc: strcpy(buf,"parameters for fractal image"); break;

	//case CHUNK_MHDR: strcpy(buf,"MNG file header"); break;
	case CHUNK_MHDR: describe_MHDR(buf); break;
	case CHUNK_MEND: strcpy(buf,"end-of-file marker"); break;
	case CHUNK_LOOP: strcpy(buf,"start of loop"); break;
	case CHUNK_ENDL: strcpy(buf,"end of loop"); break;
	case CHUNK_DEFI: strcpy(buf,"define an object"); break;
	case CHUNK_JHDR: strcpy(buf,"JNG header"); break;
	case CHUNK_BASI: strcpy(buf,"PNG chunks"); break;
	case CHUNK_CLON: strcpy(buf,"clone an object"); break;
	case CHUNK_DHDR: strcpy(buf,"Delta-PNG datastream header"); break;
	case CHUNK_PAST: strcpy(buf,"paste an image into another"); break;
	case CHUNK_DISC: strcpy(buf,"discard objects"); break;
	case CHUNK_BACK: strcpy(buf,"background"); break;
	case CHUNK_FRAM: strcpy(buf,"frame definition"); break;
	case CHUNK_MOVE: strcpy(buf,"new image location"); break;
	case CHUNK_CLIP: strcpy(buf,"object clipping boundaries"); break;
	case CHUNK_SHOW: strcpy(buf,"show images"); break;
	case CHUNK_TERM: strcpy(buf,"termination action"); break;
	case CHUNK_SAVE: strcpy(buf,"save information"); break;
	case CHUNK_SEEK: strcpy(buf,"seek point"); break;
	case CHUNK_eXPI: strcpy(buf,"export image"); break;
	case CHUNK_fPRI: strcpy(buf,"frame priority"); break;
	case CHUNK_nEED: strcpy(buf,"resources needed"); break;
	case CHUNK_pHYg: strcpy(buf,"physical pixel size (global)"); break;
	case CHUNK_JDAT: strcpy(buf,"JNG image data"); break;
	case CHUNK_JSEP: strcpy(buf,"8-bit/12-bit image separator"); break;
	case CHUNK_PROM: strcpy(buf,"promotion of parent object"); break;
	case CHUNK_IPNG: strcpy(buf,"incomplete PNG"); break;
	case CHUNK_PPLT: strcpy(buf,"partial palette"); break;
	case CHUNK_IJNG: strcpy(buf,"incomplete JNG"); break;
	case CHUNK_DROP: strcpy(buf,"drop chunks"); break;
	case CHUNK_DBYK: strcpy(buf,"drop chunks by keyword"); break;
	case CHUNK_ORDR: strcpy(buf,"ordering restrictions"); break;

	default:
		strcpy(buf,"unrecognized chunk type");
		break;
	}
}


void Chunk::get_text_descr_generic(char *buf)
{
	strcpy(buf,"");
	if(is_critical()) strcat(buf,"critical");
	else strcat(buf,"ancillary");

	if(!is_critical()) {  // safe-to-copy is not used for critical chunks
		if(is_safe_to_copy()) strcat(buf,", safe to copy");
		else strcat(buf,", unsafe to copy");
	}
}

// Several chunk types start with a null-terminated keyword. In lieu of
// fully handling them, we'll at least display that keyword.
void Chunk::describe_keyword_chunk(char *buf, int ct)
{
	int p,i;
	char *cn;

	// find null character that marks the end of the keyword
	p= -1;
	for(i=0;i<(int)length;i++) {
		if(data[i]==0) { p=i; break;}
	}

	switch(ct) {
	case CHUNK_sPLT: cn="suggested palette"; break;
	case CHUNK_iCCP: cn="embedded ICC profile"; break;
	case CHUNK_iTXt: cn="international text"; break;
	default: cn="*internal error*"; break;
	}

	if(p>=1 && p<=79) {
		sprintf(buf,"%s: [%s]",cn,data);
	}
	else {
		sprintf(buf,"%s, invalid name length",cn);
	}
}

// if exp is nonzero, does not write the length or crc
int Chunk::write_to_file(HANDLE fh, int exp)
{
	unsigned char buf[8];
	DWORD written;

	if(!exp) {
		// write length
		write_int32(&buf[0],length);
		WriteFile(fh,(LPVOID)buf,4,&written,NULL);
		if(written!=4) return 0;
	}

	// write type
	memcpy(&buf[0],chunktype,4);
	WriteFile(fh,(LPVOID)buf,4,&written,NULL);
	if(written!=4) return 0;

	// write data
	if(length>0) {
		WriteFile(fh,(LPVOID)data,length,&written,NULL);
		if(written!=length) return 0;
	}

	if(!exp) {
		// write crc
		write_int32(&buf[0],crc);
		WriteFile(fh,(LPVOID)buf,4,&written,NULL);
		if(written!=4) return 0;
	}

	return 1;
}

Chunk::Chunk(void)
{
	data=NULL;
	uncompressed_data=NULL;
	uncompressed_data_tried=0;
}

Chunk::~Chunk(void)
{
	if(data) free(data);
	if(uncompressed_data) free(uncompressed_data);
}

// call this after setting the data for the chunk
// it may do some other initialization here..
void Chunk::after_init(void)
{
	// set type id for convenience when testing chunk types
	chunktype_id=get_chunk_type_id();

	if(chunktype_id == CHUNK_IHDR) {
		if(length>=13) {
			parentpng->width=read_int32(&data[0]);
			parentpng->height=read_int32(&data[4]);
			parentpng->bitdepth=data[8];
			parentpng->colortype=data[9];
			parentpng->compression=data[10];
			parentpng->compressionfilter=data[11];
			parentpng->interlace=data[12];
		}
	}
}


// a useful function that Microsoft seems to have left out of the API
static void GetPosInParent(HWND hwnd,RECT *rc)
{
	POINT p;

	p.x= p.y= 0;
	ClientToScreen(GetParent(hwnd),&p);
	GetWindowRect(hwnd,rc);
	rc->left -= p.x;
	rc->right -= p.x;
	rc->top -= p.y;
	rc->bottom -= p.y;
}

// The text editor is so different from the other chunk editors that I'll
// keep it in a separate function from the rest.
static BOOL CALLBACK DlgProcEdit_tEXt(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	static Chunk *ch=NULL;
	int i;
	int rv;
	unsigned char *keyword, *cooked_data;

	static int border_buttonoffset=0;
	static int border_editx=0;
	static int border_edity=0;
	static int border_btn1y=0;
	static int border_btn2y=0;
	RECT rd,r1;

	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:

		SendMessage(hwnd,WM_SETICON,ICON_BIG,(LPARAM)LoadIcon(globals.hInst,"ICONMAIN"));

		ch = (Chunk*)lParam;

		if(!ch) return 1;

#ifndef TWPNG_HAVE_ZLIB
		EnableWindow(GetDlgItem(hwnd,IDC_TEXTCOMPRESSED),FALSE);
#endif

		for(i=0;known_text_keys[i];i++) {
			SendDlgItemMessage(hwnd,IDC_TEXTKEYWORD,CB_ADDSTRING,0,
			     (LPARAM)known_text_keys[i]);
		}
		SendDlgItemMessage(hwnd,IDC_TEXTKEYWORD,CB_LIMITTEXT,79,0);

		rv=ch->get_text_info(&keyword);
		if(rv) {
			SetDlgItemText(hwnd,IDC_TEXTKEYWORD,(char*)keyword);

			if(ch->uncompressed_data) {
				lf2crlf(&cooked_data,ch->uncompressed_data,ch->uncompressed_data_size);
				if(cooked_data) {
					SetDlgItemText(hwnd,IDC_TEXTTEXT,(char*)cooked_data);
					free(cooked_data);
				}
			}
		}
		CheckDlgButton(hwnd,IDC_TEXTCOMPRESSED,
			(ch->chunktype_id == CHUNK_zTXt)?BST_CHECKED:BST_UNCHECKED);


		// save some control position metrics for later resizing
		GetClientRect(hwnd,&rd);
		GetPosInParent(GetDlgItem(hwnd,IDOK),&r1);

		border_buttonoffset = (rd.right-rd.left)-r1.left;
		border_btn1y = r1.top;

		GetPosInParent(GetDlgItem(hwnd,IDCANCEL),&r1);
		border_btn2y = r1.top;

		GetPosInParent(GetDlgItem(hwnd,IDC_TEXTTEXT),&r1);
		border_editx = (rd.right-rd.left)-r1.right;
		border_edity = (rd.bottom-rd.top)-r1.bottom;

		SetWindowPos(hwnd,NULL,
			window_prefs.text.x,window_prefs.text.y,
			window_prefs.text.w,window_prefs.text.h,SWP_NOZORDER);
		if(window_prefs.text.max) ShowWindow(hwnd,SW_SHOWMAXIMIZED);

		return 1;

	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO mm;
			mm=(LPMINMAXINFO)lParam;
			mm->ptMinTrackSize.x=350;
			mm->ptMinTrackSize.y=150;
			return 1;
		}

	case WM_SIZE:
		{
			HWND h;

			GetClientRect(hwnd,&rd);

			SetWindowPos(GetDlgItem(hwnd,IDOK),NULL,rd.right-border_buttonoffset,border_btn1y,0,0,SWP_NOSIZE|SWP_NOZORDER);
			SetWindowPos(GetDlgItem(hwnd,IDCANCEL),NULL,rd.right-border_buttonoffset,border_btn2y,0,0,SWP_NOSIZE|SWP_NOZORDER);

			h=GetDlgItem(hwnd,IDC_TEXTTEXT);
			GetPosInParent(h,&r1);
			SetWindowPos(h,NULL,0,0,
				(rd.right-rd.left)-r1.left-border_editx,
				(rd.bottom-rd.top)-r1.top-border_edity,
				SWP_NOMOVE|SWP_NOZORDER);

		}
		return 1;

	case WM_DESTROY:
		StoreWindowPos(hwnd,&window_prefs.text);
		break;

	case WM_COMMAND:
		switch(id) {

		case IDOK:
			{
				HWND h;
				int len;
				unsigned char *tmptext;
				unsigned char keyword[80];
				int ct;

				ct = (IsDlgButtonChecked(hwnd,IDC_TEXTCOMPRESSED)==BST_CHECKED)
					?CHUNK_zTXt:CHUNK_tEXt;

				h=GetDlgItem(hwnd,IDC_TEXTTEXT);
				len=GetWindowTextLength(h);
				tmptext=(unsigned char*)malloc(len+1);
				GetWindowText(h,(char*)tmptext,len+1);

				// I'm not sure how to write zTXt chunks for zero-length
				// text, so I won't let the user make one.
				if(strlen((char*)tmptext)<1) {
					ct=CHUNK_tEXt;
				}

				crlf2lf(tmptext);

				GetDlgItemText(hwnd,IDC_TEXTKEYWORD,(char*)keyword,80);
				keyword[79]='\0';

				ch->set_text_info(ct,keyword,tmptext);
				free(tmptext);

				EndDialog(hwnd,1);
				return 1;
			}
				
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}


// sets tIME dialog fields to the current system time
static void set_to_current_time(HWND hwnd)
{
	SYSTEMTIME st;
	GetSystemTime(&st);
	char buf[20];

	SetDlgItemInt(hwnd,IDC_YEAR  ,(int)st.wYear,FALSE);
	SetDlgItemInt(hwnd,IDC_DAY   ,(int)st.wDay,FALSE);
	sprintf(buf,"%02d",(int)st.wHour);   SetDlgItemText(hwnd,IDC_HOUR  ,buf);
	sprintf(buf,"%02d",(int)st.wMinute); SetDlgItemText(hwnd,IDC_MINUTE,buf);
	sprintf(buf,"%02d",(int)st.wSecond); SetDlgItemText(hwnd,IDC_SECOND,buf);
	SendDlgItemMessage(hwnd,IDC_MONTH,CB_SETCURSEL,(WPARAM)((int)st.wMonth-1),0);
}

static void phys_calc_inches(HWND hwnd)
{
	int x, y;
	double xinches, yinches;
	char buf[200];
	//radio1=meters. If radio1 is checked, show in inches

	x=(IsDlgButtonChecked(hwnd,IDC_RADIO1)==BST_CHECKED)?1:0;
	if(!x) {
		SetDlgItemText(hwnd,IDC_ININCHES,"");
		return;
	}

	x=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
	y=GetDlgItemInt(hwnd,IDC_EDIT2,NULL,FALSE);

	xinches = ((double)x)*0.0254;
	yinches = ((double)y)*0.0254;
	sprintf(buf,"= %.2f x %.2f pixels/inch",xinches,yinches);
	SetDlgItemText(hwnd,IDC_ININCHES,buf);
}


// A big ugly function to handle messages for nearly all chunk edit
// dialog boxes. But the alternative of separate functions for each chunk
// type is even uglier IMHO.
static BOOL CALLBACK DlgProcEditChunk(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static Chunk *ch=NULL;
	WORD id,code;
	double tmpdbl;
	int tmpint1,tmpint2,i;
	char buf[500];

	id=LOWORD(wParam);
	code=HIWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		ch = (Chunk*)lParam;

		if(!ch) return 1;

		switch(ch->chunktype_id) {
		case CHUNK_IHDR:
			SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_ADDSTRING,0,(LPARAM)"0 = grayscale");
			SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_ADDSTRING,0,(LPARAM)"2 = RGB");
			SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_ADDSTRING,0,(LPARAM)"3 = paletted");
			SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_ADDSTRING,0,(LPARAM)"4 = grayscale+alpha");
			SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_ADDSTRING,0,(LPARAM)"6 = RGB+alpha");
			SetDlgItemInt(hwnd,IDC_EDIT1,read_int32(&ch->data[0]),FALSE);
			SetDlgItemInt(hwnd,IDC_EDIT2,read_int32(&ch->data[4]),FALSE);
			SetDlgItemInt(hwnd,IDC_EDIT3,(int)(ch->data[8]),FALSE); // bit depth
			CheckDlgButton(hwnd,IDC_CHECK1,(ch->data[12]>0)?BST_CHECKED:BST_UNCHECKED); // interlaced
			tmpint1= (int)(ch->data[9]);
			switch(tmpint1) {
			case 2: tmpint2=1; break;
			case 3: tmpint2=2; break;
			case 4: tmpint2=3; break;
			case 6: tmpint2=4; break;
			default: tmpint2=0;
			}
			SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_SETCURSEL,(WPARAM)tmpint2,0);
			break;

		case CHUNK_gAMA:
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.40000");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.45000");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.45455");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.50000");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.55000");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.60000");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.66667");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"0.70000");
			SendDlgItemMessage(hwnd,IDC_GAMMA,CB_ADDSTRING,0,(LPARAM)"1.00000");

			sprintf(buf,"%0.5f",((double)read_int32(ch->data))/100000.0);
			SetDlgItemText(hwnd,IDC_GAMMA,buf);
			break;
		case CHUNK_sRGB:
			CheckDlgButton(hwnd,IDC_SRGB_0,(ch->data[0]==0)?BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(hwnd,IDC_SRGB_1,(ch->data[0]==1)?BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(hwnd,IDC_SRGB_2,(ch->data[0]==2)?BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(hwnd,IDC_SRGB_3,(ch->data[0]==3)?BST_CHECKED:BST_UNCHECKED);
			break;

		case CHUNK_pHYs:
			SetDlgItemInt(hwnd,IDC_EDIT1,read_int32(&ch->data[0]),FALSE);
			SetDlgItemInt(hwnd,IDC_EDIT2,read_int32(&ch->data[4]),FALSE);
			CheckDlgButton(hwnd,IDC_RADIO0,(ch->data[8]==0)?BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(hwnd,IDC_RADIO1,(ch->data[8]==1)?BST_CHECKED:BST_UNCHECKED);
			phys_calc_inches(hwnd);
			break;

		case CHUNK_cHRM:
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[ 0]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT1,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[ 4]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT2,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[ 8]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT3,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[12]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT4,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[16]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT5,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[20]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT6,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[24]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT7,buf);
			sprintf(buf,"%0.5f",((double)read_int32(&ch->data[28]))/100000.0); SetDlgItemText(hwnd,IDC_EDIT8,buf);
			break;

		case CHUNK_bKGD:
			SetWindowText(hwnd,"Background color");
			tmpint1= (int)ch->parentpng->bitdepth;

			switch(ch->parentpng->colortype) {
			case 3:
				sprintf(buf,"Enter palette index (0-%d)",(1<<tmpint1)-1);
				SetDlgItemText(hwnd,IDC_LABEL1,buf);
				SetDlgItemInt(hwnd,IDC_EDIT1,(int)(ch->data[0]),FALSE);
				break;
			case 0: case 4:
				sprintf(buf,"Enter background gray shade (0-%d)",(1<<tmpint1)-1);
				SetDlgItemText(hwnd,IDC_LABEL1,buf);
				SetDlgItemInt(hwnd,IDC_EDIT1,read_int16(&ch->data[0]),FALSE);
				break;
			case 2: case 6:
				sprintf(buf,"Enter background RGB values (0-%d)",(1<<tmpint1)-1);
				SetDlgItemText(hwnd,IDC_LABEL1,buf);
				SetDlgItemInt(hwnd,IDC_EDIT1,read_int16(&ch->data[0]),FALSE);
				SetDlgItemInt(hwnd,IDC_EDIT2,read_int16(&ch->data[2]),FALSE);
				SetDlgItemInt(hwnd,IDC_EDIT3,read_int16(&ch->data[4]),FALSE);
				break;
			}
			break;

		case CHUNK_sBIT:
			SetWindowText(hwnd,"Significant bits");
			tmpint1= (ch->parentpng->colortype==3)?8:(int)ch->parentpng->bitdepth;
			sprintf(buf,"Enter significant bits (1-%d)",tmpint1);
			SetDlgItemText(hwnd,IDC_LABEL1,buf);

			if(ch->length>=1)
				SetDlgItemInt(hwnd,IDC_EDIT1,(int)(ch->data[0]),FALSE);
			if(ch->length>=2)
				SetDlgItemInt(hwnd,IDC_EDIT2,(int)(ch->data[1]),FALSE);
			if(ch->length>=3)
				SetDlgItemInt(hwnd,IDC_EDIT3,(int)(ch->data[2]),FALSE);
			if(ch->length>=4)
				SetDlgItemInt(hwnd,IDC_EDIT4,(int)(ch->data[3]),FALSE);

			break;

		case CHUNK_tRNS:
			SetWindowText(hwnd,"Transparent color");
			tmpint1= (int)ch->parentpng->bitdepth;

			switch(ch->parentpng->colortype) {
			case 0: case 4:
				sprintf(buf,"Enter transparent gray shade (0-%d)",(1<<tmpint1)-1);
				SetDlgItemText(hwnd,IDC_LABEL1,buf);
				SetDlgItemInt(hwnd,IDC_EDIT1,read_int16(&ch->data[0]),FALSE);
				break;
			case 2: case 6:
				sprintf(buf,"Enter transparent RGB color (0-%d)",(1<<tmpint1)-1);
				SetDlgItemText(hwnd,IDC_LABEL1,buf);
				SetDlgItemInt(hwnd,IDC_EDIT1,read_int16(&ch->data[0]),FALSE);
				SetDlgItemInt(hwnd,IDC_EDIT2,read_int16(&ch->data[2]),FALSE);
				SetDlgItemInt(hwnd,IDC_EDIT3,read_int16(&ch->data[4]),FALSE);
				break;
			}
			break;


		case CHUNK_tIME:
			SetDlgItemInt(hwnd,IDC_YEAR,read_int16(&ch->data[0]),FALSE);
			SetDlgItemInt(hwnd,IDC_DAY   ,(int)ch->data[3],FALSE);
			sprintf(buf,"%02d",(int)ch->data[4]); SetDlgItemText(hwnd,IDC_HOUR  ,buf);
			sprintf(buf,"%02d",(int)ch->data[5]); SetDlgItemText(hwnd,IDC_MINUTE,buf);
			sprintf(buf,"%02d",(int)ch->data[6]); SetDlgItemText(hwnd,IDC_SECOND,buf);

			for(i=1;i<=12;i++) {
				SendDlgItemMessage(hwnd,IDC_MONTH,CB_ADDSTRING,0,(LPARAM)monthname[i]);
			}
			SendDlgItemMessage(hwnd,IDC_MONTH,CB_SETCURSEL,(WPARAM)((int)ch->data[2]-1),0);
			//set_to_current_time(hwnd);

			break;
		}
		return 1;

	case WM_DESTROY:
		ch=NULL;
		break;;
		
	case WM_COMMAND:
		if(!ch) return 1;

		switch(id) {

		case IDC_EDIT1:
		case IDC_EDIT2:
			if(code==EN_CHANGE && ch->chunktype_id==CHUNK_pHYs) {
				phys_calc_inches(hwnd);
			}
			return 1;

		case IDC_RADIO0:
		case IDC_RADIO1:
			if(code==BN_CLICKED && ch->chunktype_id==CHUNK_pHYs) {
				phys_calc_inches(hwnd);
			}
			return 1;

		case IDC_BUTTON1:
			switch(ch->chunktype_id) {
			case CHUNK_tIME:
				set_to_current_time(hwnd);
				break;
			}
			return 1;

		case IDOK:
			switch(ch->chunktype_id) {

			case CHUNK_IHDR:
				tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
				tmpint2=GetDlgItemInt(hwnd,IDC_EDIT2,NULL,FALSE);
				write_int32(&ch->data[0],tmpint1); // width
				write_int32(&ch->data[4],tmpint2); // height
				ch->data[8]=(unsigned char)GetDlgItemInt(hwnd,IDC_EDIT3,NULL,FALSE); // bit depth
				ch->data[12]=(IsDlgButtonChecked(hwnd,IDC_CHECK1)==BST_CHECKED)?1:0; // interlace
				tmpint2=SendDlgItemMessage(hwnd,IDC_COLORTYPE,CB_GETCURSEL,0,0);
				switch(tmpint2) {
				case 1: tmpint1=2; break;
				case 2: tmpint1=3; break;
				case 3: tmpint1=4; break;
				case 4: tmpint1=6; break;
				default: tmpint1=0;
				}
				ch->data[9]=(unsigned char)tmpint1;
				break;

			case CHUNK_gAMA:
				GetDlgItemText(hwnd,IDC_GAMMA,buf,100);
				tmpdbl=atof(buf)*100000.0+0.000001;
				write_int32(ch->data,(DWORD)tmpdbl);
				break;

			case CHUNK_cHRM:
				GetDlgItemText(hwnd,IDC_EDIT1,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[ 0],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT2,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[ 4],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT3,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[ 8],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT4,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[12],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT5,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[16],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT6,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[20],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT7,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[24],(DWORD)tmpdbl);
				GetDlgItemText(hwnd,IDC_EDIT8,buf,100); tmpdbl=atof(buf)*100000.0+0.000001; write_int32(&ch->data[28],(DWORD)tmpdbl);
				break;

			case CHUNK_sRGB:
				if(IsDlgButtonChecked(hwnd,IDC_SRGB_0)==BST_CHECKED) ch->data[0]=0;
				if(IsDlgButtonChecked(hwnd,IDC_SRGB_1)==BST_CHECKED) ch->data[0]=1;
				if(IsDlgButtonChecked(hwnd,IDC_SRGB_2)==BST_CHECKED) ch->data[0]=2;
				if(IsDlgButtonChecked(hwnd,IDC_SRGB_3)==BST_CHECKED) ch->data[0]=3;
				break;

			case CHUNK_pHYs:
				tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
				tmpint2=GetDlgItemInt(hwnd,IDC_EDIT2,NULL,FALSE);
				if(IsDlgButtonChecked(hwnd,IDC_RADIO0)==BST_CHECKED) ch->data[8]=0;
				if(IsDlgButtonChecked(hwnd,IDC_RADIO1)==BST_CHECKED) ch->data[8]=1;
				//if(IsDlgButtonChecked(hwnd,IDC_RADIO2)==BST_CHECKED) {
				//	// convert inches to meters
				//	ch->data[8]=1;
				//	tmpint1 = (int)( ((double)tmpint1 * 100.0) /2.54);
				//	tmpint2 = (int)( ((double)tmpint2 * 100.0) /2.54);
				//}

				write_int32(&ch->data[0],tmpint1);
				write_int32(&ch->data[4],tmpint2);
				break;

			case CHUNK_sBIT:
				if(ch->length>=1)
					ch->data[0]=(unsigned char)GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
				if(ch->length>=2)
					ch->data[1]=(unsigned char)GetDlgItemInt(hwnd,IDC_EDIT2,NULL,FALSE);
				if(ch->length>=3)
					ch->data[2]=(unsigned char)GetDlgItemInt(hwnd,IDC_EDIT3,NULL,FALSE);
				if(ch->length>=4)
					ch->data[3]=(unsigned char)GetDlgItemInt(hwnd,IDC_EDIT4,NULL,FALSE);
				break;

			case CHUNK_bKGD:
				switch(ch->parentpng->colortype) {
				case 3:
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
					ch->data[0] = (unsigned char)tmpint1;
					break;
				case 0: case 4:
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
					write_int16(&ch->data[0],tmpint1);
					break;
				case 2: case 6:
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
					write_int16(&ch->data[0],tmpint1);
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT2,NULL,FALSE);
					write_int16(&ch->data[2],tmpint1);
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT3,NULL,FALSE);
					write_int16(&ch->data[4],tmpint1);
					break;
				}
				break;

			case CHUNK_tRNS:
				switch(ch->parentpng->colortype) {
				case 0: case 4:
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
					write_int16(&ch->data[0],tmpint1);
					break;
				case 2: case 6:
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,FALSE);
					write_int16(&ch->data[0],tmpint1);
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT2,NULL,FALSE);
					write_int16(&ch->data[2],tmpint1);
					tmpint1=GetDlgItemInt(hwnd,IDC_EDIT3,NULL,FALSE);
					write_int16(&ch->data[4],tmpint1);
					break;
				}
				break;

			case CHUNK_tIME:
				tmpint1=GetDlgItemInt(hwnd,IDC_YEAR,NULL,FALSE);
				write_int16(&ch->data[0],tmpint1);
				tmpint1=SendDlgItemMessage(hwnd,IDC_MONTH,CB_GETCURSEL,0,0);
				ch->data[2]=(unsigned char)(tmpint1+1);
				tmpint1=GetDlgItemInt(hwnd,IDC_DAY,NULL,FALSE);
				ch->data[3]=(unsigned char)tmpint1;
				tmpint1=GetDlgItemInt(hwnd,IDC_HOUR,NULL,FALSE);
				ch->data[4]=(unsigned char)tmpint1;
				tmpint1=GetDlgItemInt(hwnd,IDC_MINUTE,NULL,FALSE);
				ch->data[5]=(unsigned char)tmpint1;
				tmpint1=GetDlgItemInt(hwnd,IDC_SECOND,NULL,FALSE);
				ch->data[6]=(unsigned char)tmpint1;
				break;
			}
			EndDialog(hwnd,1);
			return 1;
				
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

#define EDITPAL_MARGIN 3

// fixme  Really shouldn't need this global but I was having trouble
// making variables available to both window functions.
//static palette_info_struct_t * g_editpal_palinfo_pointer;


// set i to -1 to clear values
static void update_irgba(HWND hwnd,int i,int r,int g,int b,int a,int aflag, int force)
{
	static int curr_i= -1;
	char buf_i[20],buf_r[20],buf_g[20],buf_b[20],buf_a[20];

	if(i==curr_i && !force) return;
	if(i<0) i= -1;
	curr_i=i;

	if(i<0) {
		sprintf(buf_i,"I:");
		sprintf(buf_r,"R:");
		sprintf(buf_g,"G:");
		sprintf(buf_b,"B:");
		sprintf(buf_a,"A:");
	}
	else {
		sprintf(buf_i,"I: %d",i);
		sprintf(buf_r,"R: %d",r);
		sprintf(buf_g,"G: %d",g);
		sprintf(buf_b,"B: %d",b);
		if(aflag)
			sprintf(buf_a,"A: (255)");
		else
			sprintf(buf_a,"A: %d",a);
	}

	SetDlgItemText(hwnd,IDC_STATIC_I,buf_i);
	SetDlgItemText(hwnd,IDC_STATIC_R,buf_r);
	SetDlgItemText(hwnd,IDC_STATIC_G,buf_g);
	SetDlgItemText(hwnd,IDC_STATIC_B,buf_b);
	SetDlgItemText(hwnd,IDC_STATIC_A,buf_a);
}

static void update_irgba_byindex(palette_info_struct_t *p, HWND hwnd, int c, int force)
{
	int tmp_a, tmp_aflag;
	if(p->caneditcolors) {
		tmp_a = p->plte[c].alpha;
		tmp_aflag = (c>=p->numtrns);
	}
	else {
		if(p->numtrns && c==p->trnscolor) {
			tmp_a=0;
			tmp_aflag=0;
		}
		else {
			tmp_a=255;
			tmp_aflag=1;
		}
	}
	update_irgba(hwnd,c,p->plte[c].red,p->plte[c].green,p->plte[c].blue,
		tmp_a,tmp_aflag,force);
}


// return 0-255 palette entry, or -1 for nowhere
static int where_is_cursor(int x,int y, int item_w, int item_h, int maxval)
{
	int x1,y1, v;
	if(item_w<1 || item_h<1) return -1;

	if(x<EDITPAL_MARGIN) return -1;
	if(y<EDITPAL_MARGIN) return -1;

	x1=(x-EDITPAL_MARGIN)/item_w;
	y1=(y-EDITPAL_MARGIN)/item_h;

	if(x1>15) return -1;
	if(y1>15) return -1;

	v=y1*16+x1;
	if(v>=maxval) return -1;
	return v;
}

// This is the message handler just for the embedded
// palette window that shows the palette colors, not the dialog box.
LRESULT CALLBACK WndProcEditPal(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static palette_info_struct_t *p = NULL;
	static int item_w=1, item_h=1; // size in pixels of each box
	static HWND hwndParent;
	int c;
	unsigned char cr,cg,cb;
	char buf[100];


	switch(msg) {

	case WM_SETCURSOR:
		if(p->bkgdbuttonstate || (p->alphabuttonstate && !p->caneditcolors)) {
			SetCursor(LoadCursor(NULL,IDC_CROSS));
			return 0;
		}
		break;

	case WM_CREATE:
		{
			LPCREATESTRUCT lpcs;
			lpcs = (LPCREATESTRUCT)lParam;
			p = (palette_info_struct_t*) lpcs->lpCreateParams;
			hwndParent= GetParent(hwnd);
		}
		return 0;

	case WM_MOUSEMOVE:
		c= where_is_cursor((int)(LOWORD(lParam)),(int)(HIWORD(lParam)),
			item_w,item_h, p->numplte);
		if(c<0) {
			update_irgba(hwndParent,c,0,0,0,0,0,0);
		}
		else {
			update_irgba_byindex(p,hwndParent,c,0);
		}
		return 0;

	case WM_LBUTTONDOWN:

		c= where_is_cursor((int)(LOWORD(lParam)),(int)(HIWORD(lParam)),
			item_w,item_h, p->numplte);
		if(c<0) return 0;


		if(p->bkgdbuttonstate) {
			p->bkgd=c;
			if(p->caneditcolors) {
				p->bkgdbuttonstate=0;
				CheckDlgButton(p->hwnd,IDC_SETBKGD,BST_UNCHECKED);
			}
			InvalidateRect(hwnd,NULL,TRUE);
		}
		else if(p->alphabuttonstate) {
			get_int_struct_t st;

			if(p->caneditcolors) {
				if(c>=p->numtrns) return 0;
				st.value=(int)p->plte[c].alpha;
				st.label="Enter alpha (0=transparent, 255=opaque)";
				sprintf(buf,"Alpha for palette index %d",c);
				st.title=buf;
				if(DialogBoxParam(globals.hInst,"DLG_COLOR1",hwnd,
					(DLGPROC)DlgProcGetInt, (LPARAM)(&st)) )
				{
					if(st.value<0) st.value=0;
					if(st.value>255) st.value=255;
					p->plte[c].alpha=(unsigned char)st.value;
					update_irgba(hwndParent,-2,0,0,0,0,0, 0); // clear
					InvalidateRect(hwnd,NULL,TRUE);
				}
			}
			else {
				p->trnscolor=c;
				update_irgba_byindex(p,hwndParent,c,1);
				InvalidateRect(hwnd,NULL,TRUE);
			}

		}
		else if(p->caneditcolors) {  // normal palette color edit
			cr= p->plte[c].red;
			cg= p->plte[c].green;
			cb= p->plte[c].blue;
			if( choose_color_dialog(hwnd,&cr,&cg,&cb) ) {
				p->plte[c].red   = cr;
				p->plte[c].green = cg;
				p->plte[c].blue  = cb;
				InvalidateRect(hwnd,NULL,FALSE);
			}
		}
		return 0;

	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			RECT rect;
			HBRUSH hbrush;
			HPEN hpen1,hpen2,hpen3;
			int i,xp,yp,pos;

			if(!p) break;

			hdc = BeginPaint(hwnd,&paintStruct);

			GetClientRect(hwnd,&rect);

			item_w= (rect.right-EDITPAL_MARGIN*2)/16;
			if(p->maxplte>16) {
				item_h= (rect.bottom-EDITPAL_MARGIN*2)/16;
			}
			else {
				item_h= (rect.bottom-EDITPAL_MARGIN*2);
			}

			// indicate background palette entry
			if(p->numbkgd>0) {
				xp= (p->bkgd%16)*item_w + EDITPAL_MARGIN;
				yp= (p->bkgd/16)*item_h + EDITPAL_MARGIN;

				SelectObject(hdc,GetStockObject(NULL_PEN));
				hbrush=CreateSolidBrush(RGB(0,0,255));
				SelectObject(hdc,hbrush);
				Rectangle(hdc,xp,yp,xp+item_w+1,yp+item_h+1);

				SelectObject(hdc,GetStockObject(NULL_BRUSH));   // fixme
				DeleteObject(hbrush);
			}

			SelectObject(hdc,GetStockObject(BLACK_PEN));
			for(i=0; i<p->numplte; i++) {
				xp= (i%16)*item_w + EDITPAL_MARGIN;
				yp= (i/16)*item_h + EDITPAL_MARGIN;
				hbrush=CreateSolidBrush(RGB(p->plte[i].red,p->plte[i].green,p->plte[i].blue));
				SelectObject(hdc,hbrush);
				Rectangle(hdc,xp+1,yp+2,xp+item_w-1,yp+item_h-2);
				SelectObject(hdc,GetStockObject(NULL_BRUSH));   // fixme
				DeleteObject(hbrush);
			}

			hpen1=CreatePen(PS_SOLID,0,RGB(0,128,128)); // opaque
			hpen2=CreatePen(PS_SOLID,0,RGB(128,0,0)); // translucent
			hpen3=CreatePen(PS_SOLID,0,RGB(255,0,0)); // transparent
			for(i=0; i<p->numtrns; i++) {
				int i_index;
				unsigned char i_alpha;

				if(p->caneditcolors) {
					i_index=i;
					i_alpha=p->plte[i].alpha;
				}
				else {
					i_index=p->trnscolor;
					i_alpha=0;
				}

				xp= (i_index%16)*item_w + EDITPAL_MARGIN;
				yp= (i_index/16)*item_h + EDITPAL_MARGIN;

				pos= (int)( (double)(item_w-2.1)* ((double)i_alpha / 255.0) );

				if(i_alpha==0) SelectObject(hdc,hpen3);
				else if(i_alpha==255) SelectObject(hdc,hpen1);
				else SelectObject(hdc,hpen2);

				MoveToEx(hdc,xp+1+pos,yp+item_h-7,NULL);
				LineTo(hdc,xp+1+pos,yp+item_h+0);
			}
			SelectObject(hdc,GetStockObject(NULL_PEN));
			DeleteObject(hpen1);
			DeleteObject(hpen2);
			DeleteObject(hpen3);

			EndPaint(hwnd,&paintStruct);
			return 0;
		}

	}

	return (DefWindowProc(hwnd, msg, wParam, lParam));
}


static void set_pal_labels(HWND hwnd,palette_info_struct_t *p)
{
	char buf[256];
	if(p->caneditcolors)
		sprintf(buf,"Colors in palette (%d-%d):",p->minplte,p->maxplte);
	else sprintf(buf,"Gray shades:");
	SetDlgItemText(hwnd,IDC_STATIC1,buf);

	if(p->caneditcolors)
		sprintf(buf,"Alpha values (%d-%d):",p->mintrns,(p->mintrns)?(p->numplte):0);
	else
		sprintf(buf,"Transparent colors:");
	SetDlgItemText(hwnd,IDC_STATIC2,buf);
}


static BOOL CALLBACK DlgProcEdit_PLTE(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id, code;
	int r;
	static palette_info_struct_t *p = NULL;
	static HWND hwndEditPal = NULL;

	static int border_buttonoffset=0;
	static int border_editx=0;
	static int border_edity=0;
	static int border_btn1y=0;
	static int border_btn2y=0;
	static int color_left=0;
	static int border_width=0;
	static int border_height=0;
	static int color_top=0;

	RECT rd,r1;


	id=LOWORD(wParam);
	code=HIWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:

		p = (palette_info_struct_t*)lParam;
		SendMessage(hwnd,WM_SETICON,ICON_BIG,(LPARAM)LoadIcon(globals.hInst,"ICONMAIN"));

		// save some control position metrics for later resizing
		GetClientRect(hwnd,&rd);

		GetPosInParent(GetDlgItem(hwnd,IDC_STATIC1),&r1);
		border_width=r1.left;

		GetPosInParent(GetDlgItem(hwnd,IDC_EDIT2),&r1);
		color_top= r1.bottom + 5;

		GetPosInParent(GetDlgItem(hwnd,IDC_STATIC_A),&r1);
		color_left=r1.right+3;

		GetPosInParent(GetDlgItem(hwnd,IDOK),&r1);
		border_buttonoffset = (rd.right-rd.left)-r1.left;
		border_btn1y = r1.top;
		border_height=r1.top;

		GetPosInParent(GetDlgItem(hwnd,IDCANCEL),&r1);
		border_btn2y = r1.top;

//		GetPosInParent(hwndEditPal,&r1);
//		border_editx = (rd.right-rd.left)-r1.right;
//		border_edity = (rd.bottom-rd.top)-r1.bottom;

		//g_editpal_palinfo_pointer = p;
		hwndEditPal = CreateWindow(
			"TWEAKPNGEDITPAL","EditPal",
			WS_VISIBLE|WS_CHILD|WS_BORDER,
			color_left,color_top,rd.right-border_width-color_left,
			rd.bottom-color_top-border_height,
			hwnd,NULL,globals.hInst,(void*)p);

		set_pal_labels(hwnd,p);
		update_irgba(hwnd,-2,0,0,0,0,0,0);

		SendDlgItemMessage(hwnd,IDC_EDIT1,EM_LIMITTEXT,3,0);
		SendDlgItemMessage(hwnd,IDC_EDIT2,EM_LIMITTEXT,3,0);

		SetDlgItemInt(hwnd,IDC_EDIT1,p->numplte,TRUE);
		SetDlgItemInt(hwnd,IDC_EDIT2,p->numtrns,TRUE);
		if(!p->caneditcolors) {
			SetDlgItemText(hwnd,IDC_EDITALPHA,"Set trns"); // change button label
		}

		CheckDlgButton(hwnd,IDC_SETBKGD,p->bkgdbuttonstate?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_EDITALPHA,p->alphabuttonstate?BST_CHECKED:BST_UNCHECKED);

		EnableWindow(GetDlgItem(hwnd,IDC_EDIT1),p->caneditcolors?1:0);
		EnableWindow(GetDlgItem(hwnd,IDC_EDIT2),(p->mintrns && p->caneditcolors)?1:0);
		EnableWindow(GetDlgItem(hwnd,IDC_SETBKGD),(p->numbkgd?1:0));
		EnableWindow(GetDlgItem(hwnd,IDC_EDITALPHA),(p->mintrns?1:0));
		EnableWindow(GetDlgItem(hwnd,IDC_EDITALLALPHA),(p->mintrns && p->caneditcolors)?1:0);


		p->hwnd=hwnd;

		SetWindowPos(hwnd,NULL,
			window_prefs.plte.x,window_prefs.plte.y,
			window_prefs.plte.w,window_prefs.plte.h,SWP_NOZORDER);
		if(window_prefs.plte.max) ShowWindow(hwnd,SW_SHOWMAXIMIZED);

		return TRUE;


	case WM_SIZE:

		GetClientRect(hwnd,&rd);

		SetWindowPos(GetDlgItem(hwnd,IDOK),NULL,rd.right-border_buttonoffset,border_btn1y,0,0,SWP_NOSIZE|SWP_NOZORDER);
		SetWindowPos(GetDlgItem(hwnd,IDCANCEL),NULL,rd.right-border_buttonoffset,border_btn2y,0,0,SWP_NOSIZE|SWP_NOZORDER);

		SetWindowPos(hwndEditPal,NULL,
		color_left, color_top, 
		rd.right-border_width-color_left, rd.bottom-color_top-border_height,
			SWP_NOMOVE|SWP_NOZORDER);
		return 1;

	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO mm;
			mm=(LPMINMAXINFO)lParam;
			mm->ptMinTrackSize.x=400;
			mm->ptMinTrackSize.y=280;
			return 1;
		}

	case WM_DESTROY:
		StoreWindowPos(hwnd,&window_prefs.plte);
		break;

	case WM_COMMAND:
		switch(id) {
		case IDC_EDIT1:
			if(code==EN_CHANGE) {   // user changed palette size
				r=(int)GetDlgItemInt(hwnd,IDC_EDIT1,NULL,TRUE);
				if(r<p->minplte) r=p->minplte;
				if(r>p->maxplte) r=p->maxplte;
				p->numplte=r;
				set_pal_labels(hwnd,p);
				InvalidateRect(hwndEditPal,NULL,TRUE);
				return 1;
			}
			break;
		case IDC_EDIT2:
			if(code==EN_CHANGE) {  // user changed # of alpha values
				r=(int)GetDlgItemInt(hwnd,IDC_EDIT2,NULL,TRUE);
				if(r<p->mintrns) r=p->mintrns;
				if(r>p->maxplte) r=p->maxplte;
				p->numtrns=r;
				set_pal_labels(hwnd,p);
				InvalidateRect(hwndEditPal,NULL,TRUE);
				return 1;
			}
			break;
		case IDC_SETBKGD:
			if(code==BN_CLICKED) {
				if(p->caneditcolors) {
					p->bkgdbuttonstate= !p->bkgdbuttonstate;
					CheckDlgButton(hwnd,IDC_SETBKGD,p->bkgdbuttonstate?BST_CHECKED:BST_UNCHECKED);
				}
				else {
					// If we can't edit colors, BKGD and TRNS buttons act like radio buttons.
					p->bkgdbuttonstate=1;
					p->alphabuttonstate=0;
					CheckDlgButton(hwnd,IDC_SETBKGD,BST_CHECKED);
					CheckDlgButton(hwnd,IDC_EDITALPHA,BST_UNCHECKED);
				}
			}
			break;
		case IDC_EDITALPHA:
			if(code==BN_CLICKED) {
				if(p->caneditcolors) {
					p->alphabuttonstate= !p->alphabuttonstate;
					CheckDlgButton(hwnd,IDC_EDITALPHA,p->alphabuttonstate?BST_CHECKED:BST_UNCHECKED);
				}
				else {
					p->bkgdbuttonstate=0;
					p->alphabuttonstate=1;
					CheckDlgButton(hwnd,IDC_SETBKGD,BST_UNCHECKED);
					CheckDlgButton(hwnd,IDC_EDITALPHA,BST_CHECKED);
				}
			}
			break;
		case IDC_EDITALLALPHA:
			{
				get_int_struct_t st;
				int i;

				st.value=255;
				st.label="Set all alpha (0=transparent, 255=opaque)";
				st.title="Set alpha for all palette colors";
				if(DialogBoxParam(globals.hInst,"DLG_COLOR1",hwnd,
					(DLGPROC)DlgProcGetInt, (LPARAM)(&st)) )
				{
					if(st.value<0) st.value=0;
					if(st.value>255) st.value=255;

					for(i=0;i<p->maxplte;i++) {
						p->plte[i].alpha = (unsigned char)st.value;
					}
					p->numtrns= (st.value==255)?1:p->numplte;
					SetDlgItemInt(hwnd,IDC_EDIT2,p->numtrns,TRUE);
					InvalidateRect(hwnd,NULL,TRUE);
				}
			}
			return 1;
		case IDOK:
			EndDialog(hwnd,1);
			return 1;
		case IDCANCEL:
			EndDialog(hwnd,0);
			return 1;
		}
	}
	return 0;
}


// returns 1 if successful, 0 if user canceled
static BOOL CALLBACK DlgProcGetInt(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static get_int_struct_t *st;
	WORD id;

	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		st = (get_int_struct_t*)lParam;
		SetWindowText(hwnd,st->title);
		SetDlgItemText(hwnd,IDC_LABEL1,st->label);
		SetDlgItemInt(hwnd,IDC_EDIT1,st->value,TRUE);
		return 1;

	case WM_COMMAND:
		if(!st) return 1;
		switch(id) {
		case IDOK:
			st->value=GetDlgItemInt(hwnd,IDC_EDIT1,NULL,TRUE);
			EndDialog(hwnd,1);
			return 1;
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
		break;
	}
	return 0;
}
